// Conway's Game of Life
// Implemented by Michael Cayer
// Jan 26 2005

#include <allegro.h>

#define DELAY    1         // Delay between generations.
#define ERROR    -1
#define BLACK    0
#define WHITE    15
#define FULL_X   320       // Size of the "arena" in which the automata exist.
#define FULL_Y   240

void pixelstate(BITMAP *, BITMAP *, int, int);

int main() {
  allegro_init();
  install_timer();
  install_keyboard();
  set_gfx_mode(GFX_AUTODETECT, FULL_X, FULL_Y, 0, 0);

  BITMAP *altscr = create_system_bitmap(FULL_X, FULL_Y);
  BITMAP *base = load_bitmap("life.bmp", NULL);
  if (base == NULL) {
    destroy_bitmap(altscr);
    return 1;
  }

  int x = 0;          // Used in checking cells.
  int y = 0;

  blit(base, altscr, 0, 0, 0, 0, FULL_X, FULL_Y);
  while (!keypressed()) {
    /* Display altscr. */
    blit(altscr, screen, 0, 0, 0, 0, FULL_X, FULL_Y);

    /* Read from each pixel in base successively. 
       Use pixelstate() to determine the next generation. */
    for (y = 0; y < FULL_Y; y++) {
      for (x = 0; x < FULL_X; x++) {
        pixelstate(base, altscr, x, y);
      }
    }

    /* Copy altscr to base. */
    blit(altscr, base, 0, 0, 0, 0, FULL_X, FULL_Y);
    rest(DELAY);
  }

  destroy_bitmap(base);
  destroy_bitmap(altscr);
  return 0;
}
END_OF_MAIN();

/* Pixelstate reads from the pixels adjacent to (x,y) as
   well as (x,y) itself to determine the state of a cell
   in the next generation - living or dead. */
void pixelstate(BITMAP *read, BITMAP *write, int x, int y) {
  int neighbours = 0; // Number of neighbours of a cell.

  // NORTH case.
  switch( getpixel(read, x, y-1) ) {
    case ERROR: break;
    case WHITE: break;
    case BLACK: neighbours++; break;
  }

  // NORTHEAST case.
  switch( getpixel(read, x+1, y-1) ) {
    case ERROR: break;
    case WHITE: break;
    case BLACK: neighbours++; break;
  }

  // EAST case.
  switch( getpixel(read, x+1, y) ) {
    case ERROR: break;
    case WHITE: break;
    case BLACK: neighbours++; break;
  }

  // SOUTHEAST case.
  switch( getpixel(read, x+1, y+1) ) {
    case ERROR: break;
    case WHITE: break;
    case BLACK: neighbours++; break;
  }

  // SOUTH case.
  switch( getpixel(read, x, y+1) ) {
    case ERROR: break;
    case WHITE: break;
    case BLACK: neighbours++; break;
  }

  // SOUTHWEST case.
  switch( getpixel(read, x-1, y+1) ) {
    case ERROR: break;
    case WHITE: break;
    case BLACK: neighbours++; break;
  }

  // WEST case.
  switch( getpixel(read, x-1, y) ) {
    case ERROR: break;
    case WHITE: break;
    case BLACK: neighbours++; break;
  }

  // NORTHWEST case.
  switch( getpixel(read, x-1, y-1) ) {
    case ERROR: break;
    case WHITE: break;
    case BLACK: neighbours++; break;
  }

  /* Based on number of neighbours, determine behaviour.
     Current behaviour is 23/3 - since 2 neighbours results in life 
     if alive and death if dead, it has no change. 3 will either 
     survive or create a birth cell, so it is always black. The
     rest will result in loneliness or overcrowding death, thus
     they are always white.

     Modifying this can produce other rules. */
  switch (neighbours) {
    case 0: putpixel(write, x, y, WHITE); break;
    case 1: putpixel(write, x, y, WHITE); break;
    case 2: break;
    case 3: putpixel(write, x, y, BLACK); break;
    case 4: putpixel(write, x, y, WHITE); break;
    case 5: putpixel(write, x, y, WHITE); break;
    case 6: putpixel(write, x, y, WHITE); break;
    case 7: putpixel(write, x, y, WHITE); break;
    case 8: putpixel(write, x, y, WHITE); break;
  }
}
